from tango import AttrWriteType

from facadedevice import Facade, local_attribute, logical_attribute


class Division1(Facade):
    A = local_attribute(dtype=float, access=AttrWriteType.READ_WRITE)

    B = local_attribute(dtype=float, access=AttrWriteType.READ_WRITE)

    @logical_attribute(dtype=float, bind=["A", "B"])
    def C(self, a, b):
        return a / b


if __name__ == "__main__":
    Division1.run_server()



from tango import AttrWriteType

from facadedevice import Facade, local_attribute


class Counter1(Facade):
    @local_attribute(
        dtype=int,
        access=AttrWriteType.READ_WRITE,
    )
    def count(self):
        return 0


if __name__ == "__main__":
    Counter1.run_server()



import os
import sys

sys.path.append(os.path.abspath("../"))

from facadedevice import __version__

project = "facadedevice"
version = __version__
author = "Vincent Michel"
copyright = "2016, MAX-IV"

master_doc = "index"
highlight_language = "python"
extensions = ["sphinx.ext.autodoc", "sphinx.ext.napoleon"]

html_theme = "sphinx_rtd_theme"

suppress_warnings = ["image.nonlocal_uri"]



"""Provide class objects for the facade device."""

from ast import literal_eval
from functools import partial

from tango import AttrWriteType
from tango.server import attribute, command, device_property

from facadedevice.graph import RestrictedNode, triplet
from facadedevice.utils import (
    attributes_from_wildcard,
    check_attribute,
    get_default_attribute_value,
    make_subcommand,
)

# Base class object


class class_object:
    """Provide a base for objects to be processed by ProxyMeta."""

    # Methods to override

    def update_class(self, key, dct):
        self.key = key

    def configure(self, device):
        pass  # pragma: no cover

    def connect(self, device):
        pass  # pragma: no cover

    # Representation

    def __repr__(self):
        key = self.key if self.key else "unnamed"
        return f"{type(self).__name__} <{key}>"


# Node object


class node_object(class_object):
    callback = None

    def __init__(self, use_current_time=False, **kwargs):
        self._use_current_time = use_current_time
        super().__init__(**kwargs)

    def notify(self, callback):
        """Use as a decorator to register a callback."""
        self.callback = callback
        return callback

    def configure(self, device):
        node = RestrictedNode(self.key, use_current_time=self._use_current_time)
        device.graph.add_node(node)
        # No user callback
        if not self.callback:
            return
        # Add user callback
        node.callbacks.append(
            partial(
                device._run_callback,
                "running user callback for",
                self.callback.__get__(device),
            )
        )

    # Binding helper

    @staticmethod
    def bind_node(device, node, bind, method, standard_aggregation=True):
        if not method:
            raise ValueError("No update method defined")
        if not bind:
            raise ValueError("No binding defined")
        # Set the binding
        aggregate = (
            device._standard_aggregation
            if standard_aggregation
            else device._custom_aggregation
        )
        func = partial(aggregate, node, method.__get__(device))
        device.graph.add_rule(node, func, bind)


# Local attribute


class local_attribute(node_object):
    """Tango attribute with event support.

    Local attributes support the standard attribute keywords.

    It can be used as a decorator to set a method providing the
    default value for the corresponding attribute.

    An event is pushed when the attribute is written to. That event's
    timestamp will match the time of writing. However, subsequent
    reads of the attribute will report the current timestamp (by default).
    This behaviour can be changed with an initialisation parameter.

    Args:
        create_attribute (:obj:`bool`, optional):
            Create the corresponding tango attribute. Default is True.
        use_current_time (:obj:`bool`, optional):
            Force current timestamp on each access. Default is True.
    """

    def __init__(self, create_attribute=True, use_current_time=True, **kwargs):
        if not create_attribute and kwargs:
            raise ValueError("Attribute creation is disabled")
        self.method = None
        self.kwargs = kwargs if create_attribute else None
        super().__init__(use_current_time=use_current_time)

    def __call__(self, method):
        self.method = method
        return self

    # Properties

    @property
    def use_default_write(self):
        return (
            self.kwargs is not None
            and self.kwargs.get("access") == AttrWriteType.READ_WRITE
            and not set(self.kwargs) & {"fwrite", "fset"}
        )

    # Configuration methods

    def update_class(self, key, dct):
        super().update_class(key, dct)
        # Attribute creation disabled
        if self.kwargs is None:
            return
        kwargs = dict(self.kwargs)
        # Read method
        kwargs["fget"] = lambda device, attr=None: device._read_from_node(
            device.graph[key], attr
        )
        # Is allowed method
        method_name = "is_" + key + "_allowed"
        if method_name not in dct:
            dct[method_name] = lambda device, attr: device.connected
            dct[method_name].__name__ = method_name
        # Create attribute
        dct[key] = attribute(**kwargs)
        # Read-only
        if not self.use_default_write:
            return
        # Set write method
        dct[key] = dct[key].setter(
            lambda device, value: device._write_to_node(device.graph[key], value)
        )

    def configure(self, device):
        # Build node
        super().configure(device)
        node = device.graph[self.key]
        # Attribute creation disabled
        if self.kwargs is None:
            return
        # Configure events
        attr = getattr(device, self.key)
        attr.set_archive_event(True, True)
        attr.set_change_event(True, False)
        # Add push event callback
        node.callbacks.append(
            partial(
                device._run_callback,
                "pushing events for",
                device._push_event_for_node,
            )
        )
        # Set Tango attribute value for None
        node.value_for_none = get_default_attribute_value(
            attr.get_data_format(), attr.get_data_type()
        )

    def connect(self, device):
        if not self.method:
            return
        # Get method
        node = device.graph[self.key]
        get_default = self.method.__get__(device)
        # Get default result
        try:
            result = get_default()
            if not isinstance(result, triplet):
                result = triplet(result)
        # Set exception
        except Exception as exc:
            node.set_exception(exc)
        # Set result
        else:
            node.set_result(result)


# Logical attribute


class logical_attribute(local_attribute):
    """Tango attribute computed from the values of other attributes.

    Use it as a decorator to register the function that make this computation.
    Logical attributes also support the standard attribute keywords.

    Args:
        bind (list of str):
            List of node names to bind to. It has to contain at least one name.
        standard_aggregation (:obj:`bool`, optional):
            Use the default aggregation mechanism. Default is True.
        create_attribute (:obj:`bool`, optional):
            Create the corresponding tango attribute. Default is True.
    """

    def __init__(self, bind, standard_aggregation=True, **kwargs):
        self.bind = bind
        self.method = None
        self.standard_aggregation = standard_aggregation
        super().__init__(use_current_time=False, **kwargs)

    def configure(self, device):
        super().configure(device)
        node = device.graph[self.key]
        self.configure_binding(device, node)

    def configure_binding(self, device, node):
        self.bind_node(device, node, self.bind, self.method, self.standard_aggregation)

    def connect(self, device):
        # Override the local_attribute connect method
        pass


# Proxy attribute


class proxy_attribute(logical_attribute):
    """Tango attribute linked to the attribute of a remote device.

    Args:
        property_name (str):
            Name of the property containing the attribute name.
        create_property (:obj:`bool`, optional):
            Create the corresponding device property. Default is True.
        standard_aggregation (:obj:`bool`, optional):
            Use the default aggregation mechanism. Default is True.
        create_attribute (:obj:`bool`, optional):
            Create the corresponding tango attribute. Default is True.

    Also supports the standard attribute keywords.
    """

    def __init__(self, property_name, create_property=True, **kwargs):
        self.property_name = property_name
        self.create_property = create_property
        super().__init__(None, **kwargs)

    def update_class(self, key, dct):
        # Parent method
        super().update_class(key, dct)
        # Create device property
        if self.create_property:
            doc = f"Attribute to be forwarded as {key}."
            dct[self.property_name] = device_property(dtype=str, doc=doc)
        # Read-only or custom write
        if not self.use_default_write:
            return
        # Set write method
        dct[key] = dct[key].setter(
            lambda device, value: device._run_proxy_command(key, value)
        )

    def configure_binding(self, device, node):
        # Get properties
        attr = getattr(device, self.property_name).strip()
        # Empty property
        if not attr:
            msg = "Property {!r} is empty"
            raise ValueError(msg.format(self.property_name))
        # Default value
        if "/" not in attr:
            node.default_value = literal_eval(attr)
            # Make subcommand
            if self.use_default_write:
                subcommand = partial(device._write_to_node, node)
                device._subcommand_dict[self.key] = subcommand
            return
        # Check attribute
        attr = attr.lower()
        check_attribute(attr, writable=self.use_default_write)
        # Make subcommand
        if self.use_default_write:
            subcommand = make_subcommand(attr, attr=True)
            device._subcommand_dict[self.key] = subcommand
        # Add attribute
        if self.method is None:
            node.remote_attr = attr
            return
        # Add subnode
        bind = (self.key + "[0]",)
        subnode = RestrictedNode(bind[0])
        subnode.remote_attr = attr
        device.graph.add_node(subnode)
        # Binding
        self.bind_node(device, node, bind, self.method, self.standard_aggregation)

    def connect(self, device):
        node = device.graph[self.key]
        # Set default_value
        if hasattr(node, "default_value"):
            node.set_result(triplet(node.default_value))
        # Get subnodes
        if hasattr(node, "remote_attr"):
            subnodes = [node]
        else:
            subnodes = device.graph.subnodes(self.key)
        # Subscribe
        for subnode in subnodes:
            device._subscribe_for_node(subnode.remote_attr, subnode)


# Combined attribute


class combined_attribute(proxy_attribute):
    """Tango attribute computed from the values of other remote attributes.

    Use it as a decorator to register the function that make this computation.
    The remote attribute names are provided by a property, either as a list or
    a pattern.

    Args:
        property_name (str):
            Name of the property containing the attribute names.
        create_property (:obj:`bool`, optional):
            Create the corresponding device property. Default is True.
        standard_aggregation (:obj:`bool`, optional):
            Use the default error aggregation mechanism. Default is True.
        create_attribute (:obj:`bool`, optional):
            Create the corresponding tango attribute. Default is True.

    Also supports the standard attribute keywords.
    """

    def update_class(self, key, dct):
        # Parent method
        super().update_class(key, dct)
        # Check write access
        if self.use_default_write:
            raise ValueError(f"{self} cannot be writable")
        # Override device property
        if self.create_property:
            doc = f"Attributes to be combined as {key}."
            dct[self.property_name] = device_property(dtype=(str,), doc=doc)

    def configure_binding(self, device, node):
        # Strip property
        attrs = getattr(device, self.property_name)
        if not isinstance(attrs, str):
            attrs = "\n".join(attrs)
        attrs = attrs.strip()
        # Empty property
        if not attrs:
            msg = "Property {!r} is empty"
            raise ValueError(msg.format(self.property_name))
        # Default value
        if "/" not in attrs:
            node.default_value = literal_eval(attrs)
            return
        # Split lines
        attrs = attrs.lower().splitlines()
        attrs = list(filter(None, map(str.strip, attrs)))
        # Pattern matching
        if len(attrs) == 1:
            wildcard = attrs[0]
            attrs = list(attributes_from_wildcard(wildcard))
            if not attrs:
                msg = "No attributes matching {} wildcard"
                raise ValueError(msg.format(wildcard))
        # Check attributes
        else:
            for attr in attrs:
                check_attribute(attr)
        # Build the bindings
        bind = tuple(f"{self.key}[{i}]" for i, _ in enumerate(attrs))
        # Build the subnodes
        for key, attr in zip(bind, attrs):
            subnode = RestrictedNode(key)
            subnode.remote_attr = attr
            device.graph.add_node(subnode)
        # Set the binding
        self.bind_node(device, node, bind, self.method, self.standard_aggregation)


# State attribute


class state_attribute(node_object):
    """Tango state attribute with event support.

    Args:
        bind (list of str):
            List of node names to bind to, or None to disable the binding.
            Default is None.
        standard_aggregation (:obj:`bool`, optional):
            Use the default error aggregation mechanism. Default is True.
    """

    def __init__(self, bind=None, standard_aggregation=True, **kwargs):
        self.bind = bind
        self.method = None
        self.standard_aggregation = standard_aggregation
        super().__init__(**kwargs)

    def __call__(self, method):
        self.method = method
        return self

    def update_class(self, key, dct):
        super().update_class(key, dct)
        # Restore method
        if self.method:
            self.method.bind = self.bind
            dct[key] = self.method

    def configure(self, device):
        # Parent call
        super().configure(device)
        node = device.graph[self.key]
        # Add set state callback
        node.callbacks.append(
            partial(
                device._run_callback,
                "setting the state from",
                device._set_state_from_node,
            )
        )
        # Nothing to bind
        if not self.bind and not self.method:
            return
        # Bind node
        self.bind_node(device, node, self.bind, self.method, self.standard_aggregation)


# Proxy command


class proxy_command(class_object):
    """Command to write an attribute or run a command of a remote device.

    It can be used as a decorator to define a more precise behavior.
    The decorated method takes the subcommand as its first argument.

    Args:
        property_name (str):
            Name of the property containing the attribute or command name.
        create_property (str):
            Create the corresponding device property. Default is True.
        write_attribute (bool):
            True if the subcommand should an attribute write, False otherwise.
            Default is false.

    Also supports the standard command keywords.
    """

    def __init__(
        self, property_name, create_property=True, write_attribute=False, **kwargs
    ):
        self.kwargs = kwargs
        self.property_name = property_name
        self.create_property = create_property
        self.write_attribute = write_attribute
        # Default method
        self.method = lambda device, sub, *args: sub(*args)

    def __call__(self, method):
        self.method = method
        return self

    def update_class(self, key, dct):
        super().update_class(key, dct)
        # Set command
        dct[key] = lambda device, *args: device._run_proxy_command_context(
            key, self.method.__get__(device), *args
        )
        dct[key].__name__ = key
        dct[key] = command(**self.kwargs)(dct[key])
        # Set is allowed method
        method_name = "is_" + key + "_allowed"
        if method_name not in dct:
            dct[method_name] = lambda device: device.connected
            dct[method_name].__name__ = method_name
        # Create property
        if self.create_property:
            doc = (
                "Attribute to be written"
                if self.write_attribute
                else "Subcommand to be executed"
            )
            doc += f" in {key} command."
            dct[self.property_name] = device_property(dtype=str, doc=doc)

    def configure(self, device):
        name = getattr(device, self.property_name).strip()
        # Default value
        if "/" not in name:
            value = literal_eval(name)
            subcommand = partial(device._emulate_subcommand, value)
        # Check subcommand
        else:
            subcommand = make_subcommand(name, attr=self.write_attribute)
        # Set subcommand
        device._subcommand_dict[self.key] = subcommand



from facadedevice import Facade, proxy_command


class Commands(Facade):
    reset = proxy_command(property_name="ResetCommand")

    echo = proxy_command(dtype_in=str, dtype_out=str, property_name="EchoCommand")

    set_level = proxy_command(
        dtype_in=float, property_name="LevelAttribute", write_attribute=True
    )

    @proxy_command(dtype_in=int, dtype_out=int, property_name="EchoCommand")
    def identity(self, subcommand, arg):
        return int(subcommand(str(arg)))


if __name__ == "__main__":
    Commands.run_server()



from facadedevice import Facade, logical_attribute, proxy_attribute


class Division2(Facade):
    A = proxy_attribute(dtype=float, property_name="AAttribute")

    B = proxy_attribute(dtype=float, property_name="BAttribute")

    @logical_attribute(dtype=float, bind=["A", "B"])
    def C(self, a, b):
        return a / b


if __name__ == "__main__":
    Division2.run_server()



from tango import DevState

from facadedevice import (
    Facade,
    logical_attribute,
    proxy_attribute,
    proxy_command,
    state_attribute,
)


class CameraScreen(Facade):
    # Proxy attributes

    StatusIn = proxy_attribute(dtype=bool, property_name="StatusInAttribute")

    StatusOut = proxy_attribute(dtype=bool, property_name="StatusOutAttribute")

    # Logical attributes

    @logical_attribute(dtype=bool, bind=["StatusIn", "StatusOut"])
    def Error(self, status_in, status_out):
        return status_in and status_out

    @logical_attribute(dtype=bool, bind=["StatusIn", "StatusOut"])
    def Moving(self, status_in, status_out):
        return not status_in and not status_out

    # Proxy commands

    @proxy_command(property_name="MoveInAttribute", write_attribute=True)
    def MoveIn(self, subcommand):
        subcommand(1)

    @proxy_command(property_name="MoveOutAttribute", write_attribute=True)
    def MoveOut(self, subcommand):
        subcommand(1)

    # State and status

    @state_attribute(bind=["Error", "Moving", "StatusIn"])
    def state(self, error, moving, status_in):
        if error:
            return DevState.FAULT, "A conflict has been detected"
        elif moving:
            return DevState.MOVING, "The screen is moving"
        elif status_in:
            return DevState.INSERT, "The screen is inserted"
        else:
            return DevState.EXTRACT, "The screen is exctracted"


if __name__ == "__main__":
    CameraScreen.run_server()



"""Provide objects for building reactive graphs."""

import warnings
from collections import defaultdict, namedtuple
from collections.abc import Mapping
from functools import partial

import numpy as np
from tango import AttrQuality

from facadedevice.utils import timestamp_now

# Constants

VALID = AttrQuality.ATTR_VALID
INVALID = AttrQuality.ATTR_INVALID


# Triplet object

triplet = namedtuple("triplet", ("value", "stamp", "quality"))


def new_triplet(cls, value, stamp=None, quality=VALID):
    if isinstance(value, triplet):
        raise TypeError("The value cannot be a triplet")
    if stamp is None:
        stamp = timestamp_now()
    if not isinstance(stamp, float):
        raise TypeError("The timestamp is not a float")
    if not isinstance(quality, int):
        raise TypeError("The quality is not a integer")
    if value is None or quality == INVALID:
        value, quality = None, INVALID
    return cls.__rawnew__(cls, value, stamp, quality)


def compare_triplet(a, b):
    assert isinstance(a, triplet)
    try:
        value, stamp, quality = b
    except Exception:
        return False
    return a.stamp == stamp and a.quality == quality and np.array_equal(a.value, value)


def from_attr_value(cls, attr_value):
    return cls(attr_value.value, attr_value.time.totime(), attr_value.quality)


triplet.__rawnew__ = staticmethod(triplet.__new__)
triplet.__new__ = staticmethod(new_triplet)
triplet.__eq__ = compare_triplet
triplet.__ne__ = lambda self, arg: not self.__eq__(arg)
triplet.from_attr_value = classmethod(from_attr_value)


# Node object


class Node:
    def __init__(self, name, description=None, callbacks=()):
        self._result = None
        self._exception = None
        self.name = name
        self.description = description or name
        self.callbacks = list(callbacks)
        self.value_for_none = None

    # Setters

    def set_result(self, result):
        diff = self._exception is not None or self._result != result
        self._result = result
        self._exception = None
        if diff:
            self.notify()

    def set_exception(self, exception):
        if not isinstance(exception, BaseException):
            raise TypeError("Not a valid exception")
        diff = self._result is not None or self._exception != exception
        self._result = None
        self._exception = exception
        if diff:
            self.notify()

    # Getters

    def result(self):
        if self._exception is not None:
            raise self._exception
        return self._result

    def exception(self):
        return self._exception

    # Notifying

    def notify(self):
        for callback in self.callbacks:
            try:
                callback(self)
            except Exception as exc:
                message = "Node {} failed to notify: {!r}"
                warnings.warn(message.format(self.name, exc))

    # Representation

    def __repr__(self):
        return f"node <{self.name}>"


# Restricted node


class RestrictedNode(Node):
    def __init__(self, name, description=None, callbacks=(), use_current_time=False):
        super().__init__(name, description, callbacks)
        self.use_current_time = use_current_time

    def set_result(self, result):
        if result is not None and not isinstance(result, triplet):
            raise TypeError("Not a triplet (or None)")
        return super().set_result(result)

    def result(self):
        result = super().result()
        if result is not None and self.use_current_time:
            # force timestamp to now (keep value and quality)
            return triplet(
                value=result.value, stamp=timestamp_now(), quality=result.quality
            )
        else:
            return result


# Graph object


class Graph(Mapping):
    def __init__(self):
        # Graph state
        self._nodes = {}
        self._rules = {}
        # Computed state
        self._updates = {}
        self._dependencies = defaultdict(set)
        self._subscriptions = defaultdict(set)
        # Propagation state
        self._pending = set()
        self._propagating = False

    # Accessors

    def subnodes(self, name):
        node = self._nodes[name]
        if node not in self._rules:
            return []
        _, bind = self._rules[node]
        return [self._nodes[subname] for subname in bind]

    # Create graph

    def add_node(self, node):
        if node.name in self._nodes:
            message = "A node called {} already exists"
            raise ValueError(message.format(node.name))
        self._nodes[node.name] = node

    def add_rule(self, node, func, bind):
        if node not in self._nodes.values():
            message = "The node {!r} is not in the graph"
            raise ValueError(message.format(node))
        if node in self._rules:
            message = "A rule for {} already exists"
            raise ValueError(message.format(node.name))
        self._rules[node] = func, bind

    # Build dependencies

    def build(self):
        # Loop over rules
        for node, (func, bind) in self._rules.items():
            # Set update callbacks
            publishers = [self._nodes[subname] for subname in bind]
            self._updates[node] = partial(func, *publishers)
            # Set subscriptions
            for publisher in publishers:
                self._subscriptions[publisher].add(node)
            # Compute dependencies
            seen = set()
            names = set(bind)
            while names:
                current_name = names.pop()
                seen.add(current_name)
                current_node = self._nodes[current_name]
                if current_node not in self._rules:
                    continue
                current_rule = self._rules[current_node]
                names |= set(current_rule[1]) - seen
            # Check cyclic dependencies
            if node.name in seen:
                msg = "{} is involved in a cyclic dependency"
                raise ValueError(msg.format(node))
            # Set dependencies
            self._dependencies[node] = {self._nodes[name] for name in seen}
        # Set callbacks (after the graph is proven to be valid)
        for publisher in self._subscriptions.keys():
            if self.callback not in publisher.callbacks:
                publisher.callbacks.append(self.callback)

    # Reset

    def reset(self):
        # Remove callbacks
        for publisher in self._subscriptions.keys():
            if self.callback in publisher.callbacks:
                publisher.callbacks.remove(self.callback)
        # Reset computed state
        self._updates = {}
        self._dependencies = defaultdict(set)
        self._subscriptions = defaultdict(set)
        # Reset propagation state
        self._pending = set()
        self._propagating = False

    # Propagation

    def callback(self, node):
        self._pending |= self._subscriptions[node]
        if not self._propagating:
            self.propagate()

    def propagate(self):
        # Set propagation flag
        try:
            self._propagating = True
            # Loop over pending updates
            while self._pending:
                # Find next item
                for node in self._pending:
                    dependencies = self._dependencies[node]
                    if self._pending.isdisjoint(dependencies):
                        self._pending.remove(node)
                        break
                # Deadlock
                else:
                    warnings.warn("Propagation deadlocked")
                    node = self._pending.pop()
                # Update and notify
                self.update(node)
        # Reset propagation flag
        finally:
            self._propagating = False

    def update(self, node):
        callback = self._updates[node]
        try:
            node.set_result(callback())
        except Exception as exc:
            node.set_exception(exc)

    # Dict interface

    def __getitem__(self, key):
        return self._nodes[key]

    def __iter__(self):
        return iter(self._nodes)

    def __len__(self):
        return len(self._nodes)



"""Provide helpers for tango."""

import collections
import fnmatch
import functools
import itertools
import time

try:
    from threading import get_ident
except ImportError:  # pragma: no cover
    from threading import _get_ident as get_ident

from tango import (
    AttrDataFormat,
    AttrQuality,
    AttrWriteType,
    AutoTangoMonitor,
    CmdArgType,
    Database,
    DeviceProxy,
    DevState,
    DispLevel,
)
from tango.server import Device, command

try:
    from tango import EventSubMode
except ImportError:  # pragma: no cover
    EventSubMode = None


from facadedevice.exception import exception_string, traceback_string

# Constants

ATTR_NOT_ALLOWED = "API_AttrNotAllowed"


# Default attribute value


def get_default_attribute_value(dformat, dtype):
    if dformat == AttrDataFormat.IMAGE:
        return ((),)
    if dformat == AttrDataFormat.SPECTRUM:
        return ()
    if dtype == CmdArgType.DevString:
        return ""
    # Zero should work for everything else
    return 0


# Aggregate qualities


def aggregate_qualities(qualities):
    length = len(AttrQuality.values)
    sortable = map(lambda x: (x - 1) % length, qualities)
    result = (min(sortable) + 1) % length
    return AttrQuality.values[result]


# Patched device proxy


def create_device_proxy(*args, **kwargs):
    proxy = DeviceProxy(*args, **kwargs)
    proxy._get_info_()
    return proxy


# Split attribute name


def split_tango_name(name):
    lst = name.split("/")
    return "/".join(lst[:-1]), lst[-1]


# Attribute check


def check_attribute(name, writable=False):
    device, attr = split_tango_name(name)
    proxy = create_device_proxy(device)
    cfg = proxy.get_attribute_config(attr)
    if writable and cfg.writable is AttrWriteType.READ:
        raise ValueError(f"The attribute {name} is not writable")
    return cfg


# Attribute from wildcard


def attributes_from_wildcard(wildcard):
    db = Database()
    wdev, wattr = split_tango_name(wildcard)
    for device in db.get_device_exported(wdev):
        proxy = create_device_proxy(device)
        infos = proxy.attribute_list_query()
        attrs = sorted(info.name.lower() for info in infos)
        for attr in fnmatch.filter(attrs, wattr):
            yield f"{device.lower()}/{attr}"


# Tango command check


def check_command(name):
    device, cmd = split_tango_name(name)
    proxy = create_device_proxy(device)
    return proxy.command_query(cmd)


# Make subcommand


def make_subcommand(name, attr=False):
    # Check value
    if attr:
        check_attribute(name)
    else:
        check_command(name)
    # Create proxy
    device, obj = split_tango_name(name)
    proxy = create_device_proxy(device)
    # Make subcommand
    method = proxy.write_attribute if attr else proxy.command_inout
    return functools.partial(method, obj)


# Get current time


def timestamp_now():
    return time.time()


# Device class


class EnhancedDevice(Device):
    """Enhanced version of server.Device"""

    # Property

    @property
    def connected(self):
        return self._connected

    # Exception helpers

    def register_exception(self, exc, msg=None, ignore=False):
        # Stream traceback
        self.debug_stream(traceback_string(exc).replace("%", "%%"))
        # Exception as a string
        status = exception_string(exc, wrap=msg)
        # Stream error
        self.error_stream(status)
        # Save in history
        self._exception_history[status] += 1
        # Set state and status
        if not ignore:
            self.set_status(status)
            self.set_state(DevState.FAULT)
        # Return exception status
        return status

    def ignore_exception(self, exc, msg=None):
        return self.register_exception(exc, msg=msg, ignore=True)

    def debug_exception(self, exc, msg=None):
        string = exception_string(exc, wrap=msg)
        self.debug_stream(string.replace("%", "%%"))

    # Initialization and cleanup

    def get_device_properties(self, cls=None):
        """Raise a ValueError if a property is missing."""
        Device.get_device_properties(self, cls)
        for key, value in self.device_property_list.items():
            if value[2] is None:
                raise ValueError("Missing property: " + key)

    def init_device(self):
        # Init attributes
        self._event_dict = {}
        self._connected = False
        self._tango_properties = {}
        self._init_stamp = timestamp_now()
        self._eid_counter = itertools.count(1)
        self._exception_history = collections.defaultdict(int)
        # Init state and status events
        self.set_change_event("State", True, False)
        self.set_archive_event("State", True, True)
        self.set_change_event("Status", True, False)
        self.set_archive_event("Status", True, True)
        # Set INIT state
        self.set_state(DevState.INIT)
        # Initialize the device
        try:
            self._init_ident = get_ident()
            self.safe_init_device()
        # Register exception
        except Exception as exc:
            msg = "Exception while initializing the device"
            self.register_exception(exc, msg)
            self.delete_device()
            return
        # Connection successful
        else:
            self._connected = True
            self._init_ident = None
        # Set default state
        if self.get_state() == DevState.INIT:
            self.set_state(DevState.UNKNOWN)

    def safe_init_device(self):
        self.get_device_properties()

    def delete_device(self):
        # Unsubscribe all
        try:
            self.unsubscribe_all()
        except Exception as exc:
            msg = "Exception while unsubscribing"
            return self.ignore_exception(exc, msg)
        # Parent call
        super(Device, self).delete_device()

    # Event subscribtion

    def _wrap_callback(self, callback, eid):
        def wrapped(event):
            # Fix libtango bug #316
            if self._init_ident is not None and self._init_ident != get_ident():
                return  # pragma: no cover
            # Acquire monitor lock
            try:
                with AutoTangoMonitor(self):
                    if eid in self._event_dict:
                        callback(event)
            # Register exception
            except Exception as exc:
                message = "Exception while running event callback {}"
                self.ignore_exception(exc, message.format(eid))

        return wrapped

    def subscribe_event(
        self,
        attr_name,
        event_type,
        callback,
        filters=[],
        stateless=False,
        proxy=None,
    ):
        # Get proxy
        if proxy is None:
            device_name, attr_name = split_tango_name(attr_name)
            proxy = create_device_proxy(device_name)
        # Create callback
        eid = next(self._eid_counter)
        self._event_dict[eid] = proxy, attr_name, None
        wrapped = self._wrap_callback(callback, eid)
        # Subscribe
        try:
            if EventSubMode is not None:
                # PyTango >= 10.1.0
                sub_mode = (
                    EventSubMode.Stateless if stateless else EventSubMode.SyncRead
                )
                proxy_eid = proxy.subscribe_event(
                    attr_name, event_type, wrapped, sub_mode=sub_mode
                )
            else:
                # filters and stateless params are deprecated since PyTango 10.1.0
                proxy_eid = proxy.subscribe_event(
                    attr_name, event_type, wrapped, filters, stateless
                )
        # Error
        except Exception:
            del self._event_dict[eid]
            raise
        # Success
        self._event_dict[eid] = proxy, attr_name, proxy_eid, event_type
        return eid

    def unsubscribe_event(self, eid):
        proxy, _, proxy_eid, _ = self._event_dict.pop(eid)
        proxy.unsubscribe_event(proxy_eid)

    def unsubscribe_all(self):
        for eid in list(self._event_dict):
            proxy, attr_name, _, _ = self._event_dict[eid]
            attr_name = "/".join((proxy.dev_name(), attr_name))
            try:
                self.unsubscribe_event(eid)
            except Exception as exc:
                msg = "Exception while unsubscribing from attribute {}"
                msg = msg.format(attr_name)
                self.ignore_exception(exc, msg)
            else:
                msg = "Successfully unsubscribed from attribute {}"
                msg = msg.format(attr_name)
                self.info_stream(msg)

    # State, status

    def set_state(self, state, stamp=None, quality=AttrQuality.ATTR_VALID):
        super(Device, self).set_state(state)
        # Pushing specific values for events on state attribute doesn't work
        self.push_change_event("State")  # ... state, stamp, quality)
        self.push_archive_event("State")  # ... state, stamp, quality)

    def set_status(self, status, stamp=None, quality=AttrQuality.ATTR_VALID):
        super(Device, self).set_status(status)
        # Pushing specific values for events on status attribute doesn't work
        self.push_change_event("Status")  # ... state, stamp, quality)
        self.push_archive_event("Status")  # ... state, stamp, quality)

    # Commands

    @command(
        dtype_out=str,
        doc_out="Information about events and exceptions.",
        display_level=DispLevel.EXPERT,
    )
    def GetInfo(self):
        lines = []
        # Connection
        if self._connected:
            lines.append("The device is currently connected.")
        else:
            lines.append("The device is currently stopped because of:")
            lines.append(self.get_status())
        # Event subscription
        if self._event_dict:
            lines.append(
                "It subscribed to event channel of the following attribute(s):"
            )
            for proxy, attr_name, _, event_type in self._event_dict.values():
                attr_name = "/".join((proxy.dev_name(), attr_name))
                lines.append(f"- {attr_name} ({event_type})")
        else:
            lines.append("It doesn't hold a subsription to any event channel.")
        # Exception history
        lines.append("-" * 5)
        strtime = time.ctime(self._init_stamp)
        if self._exception_history:
            msg = "Error history since {} (last initialization):"
            lines.append(msg.format(strtime))
            for key, value in self._exception_history.items():
                string = "once" if value == 1 else f"{value} times"
                lines.append(f" - Raised {string}:")
                lines.extend(" " * 4 + line for line in key.split("\n"))
        else:
            msg = "No errors in history since {} (last initialization)."
            lines.append(msg.format(strtime))
        # Return result
        return "\n".join(lines)



from tango.server import command

from facadedevice import Facade, local_attribute, triplet


class Counter2(Facade):
    @local_attribute(dtype=int)
    def count(self):
        return 0

    @command
    def increment(self):
        node = self.graph["count"]
        value, stamp, quality = node.result()
        new_result = triplet(value + 1)
        node.set_result(new_result)


if __name__ == "__main__":
    Counter2.run_server()



"""Provide the facade device class and metaclass."""

import collections

from tango import DevFailed, DevState, DispLevel, EventData, EventType
from tango.server import command

from facadedevice.exception import context, to_dev_failed
from facadedevice.graph import INVALID, Graph, triplet
from facadedevice.objects import class_object, local_attribute
from facadedevice.utils import (
    EnhancedDevice,
    aggregate_qualities,
    timestamp_now,
)

# Proxy metaclass


class FacadeMeta(type(EnhancedDevice)):
    """Metaclass for Facade device."""

    def __new__(metacls, name, bases, dct):
        # Class attribute
        dct["_class_dict"] = class_dict = {}
        # Inheritance
        for base in reversed(bases):
            try:
                base_class_dict = base._class_dict
            except AttributeError:
                continue
            # Copy _class_dict from the bases
            for key, obj in base_class_dict.items():
                # Allow to remove facade objects by setting them to None
                if key not in dct:
                    class_dict[key] = obj
        # Process class objects
        for key, value in list(dct.items()):
            if isinstance(value, class_object):
                class_dict[key] = value
                value.update_class(key, dct)
        # Create device class
        return type(EnhancedDevice).__new__(metacls, name, bases, dct)


# Metaclassing manually for python compatibility

_Facade = FacadeMeta("_Facade", (EnhancedDevice,), {})


# Facade device


class Facade(_Facade):
    """Base class for facade devices.

    It supports the following objects:

    - :class:`facadedevice.local_attribute`
    - :class:`facadedevice.logical_attribute`
    - :class:`facadedevice.proxy_attribute`
    - :class:`facadedevice.combined_attribute`
    - `:class:`facadedevice.state_attribute`
    - :class:`facadedevice.proxy_command`

    It also provides a few helpers:

    - `self.graph`: act as a `<key, node>` dictionnary
    - `self.get_combined_results`: return the subresults of a combined
      attribute

    The `init_device` method shouldn't be overridden. It performs specific
    exception handling. Instead, override `safe_init_device` if you have to
    add some extra logic. Don't forget to call the parent method since it
    performs a few useful steps:

    - load device properties
    - configure and build the graph
    - run the connection routine

    It also provides an expert command called `GetInfo` that displays useful
    information such as:

    - the connection status
    - the list of all event subscriptions
    - the exception history
    """

    # Reasons to ignore for errors in events
    reasons_to_ignore = ["API_PollThreadOutOfSync"]

    # Properties

    @property
    def graph(self):
        return self._graph

    # Helper

    def get_combined_results(self, name):
        """Return the subresults of a given combined attribute.

        It produces an ordered dictionnary of <attribute_name, triplet>.
        """
        subnodes = self.graph.subnodes(name)
        return collections.OrderedDict(
            (node.remote_attr, node.result()) for node in subnodes
        )

    def _emulate_subcommand(self, result, *args):
        if args or result is None:
            raise ValueError("This proxy command is disabled")
        return result

    # Initialization

    def safe_init_device(self):
        """Initialize the device."""
        # Init data structures
        self._graph = Graph()
        self._subcommand_dict = {}
        # Get properties
        with context("getting", "properties"):
            super().safe_init_device()
        # Configure
        for value in self._class_dict.values():
            with context("configuring", value):
                value.configure(self)
        # Build graph
        with context("building", self._graph):
            self._graph.build()
        # Connect
        for value in self._class_dict.values():
            with context("connecting", value):
                value.connect(self)

    # Event subscription

    def _subscribe_for_node(self, attr, node):
        try:
            self.subscribe_event(
                attr,
                EventType.CHANGE_EVENT,
                lambda event: self._on_node_event(node, event),
            )
        except DevFailed:
            try:
                self.subscribe_event(
                    attr,
                    EventType.PERIODIC_EVENT,
                    lambda event: self._on_node_event(node, event),
                )
            except DevFailed:
                msg = "Can't subscribe to event for attribute {}"
                self.info_stream(msg.format(attr))
                raise
            else:
                msg = "Subscribed to periodic event for attribute {}"
                self.info_stream(msg.format(attr))
                return EventType.PERIODIC_EVENT
        else:
            msg = "Subscribed to change event for attribute {}"
            self.info_stream(msg.format(attr))
            return EventType.CHANGE_EVENT

    # Event callback

    def _on_node_event(self, node, event):
        """Handle node events."""
        # Ignore the event if not a data event
        if not isinstance(event, EventData):
            msg = "Received an unexpected event for {}"
            self.error_stream(msg.format(node))
            return
        # Format attribute name
        attr_name = "/".join(event.attr_name.split("/")[-4:])
        # Ignore the event if it contains an error
        if event.errors:
            exc = DevFailed(*event.errors)
            reason = exc.args[0].reason
            msg = "Received an event from {} that contains errors"
            self.ignore_exception(exc, msg=msg.format(attr_name))
            if reason not in self.reasons_to_ignore:
                node.set_exception(exc)
            return
        # Info stream
        msg = "Received a valid event from {} for {}."
        self.info_stream(msg.format(attr_name, node))
        # Save
        value = triplet.from_attr_value(event.attr_value)
        node.set_result(value)

    # Client requests

    def _read_from_node(self, node, attr=None):
        """Used when reading an attribute"""
        result = node.result()
        if result is None:
            return
        value, stamp, quality = result
        if value is None:
            value = node.value_for_none
        subnodes = self.graph.subnodes(node.name)
        force_current_time = any(subnode.use_current_time for subnode in subnodes)
        if force_current_time:
            stamp = timestamp_now()
        if attr:
            attr.set_value_date_quality(value, stamp, quality)
        return value, stamp, quality

    def _write_to_node(self, node, value):
        """Used when writing a local attribute"""
        node.set_result(triplet(value))

    def _run_proxy_command(self, key, value):
        """Used when writing a proxy attribute"""
        return self._run_proxy_command_context(
            key, lambda subcommand, value: subcommand(value), value
        )

    def _run_proxy_command_context(self, key, ctx, *values):
        """Used when running a proxy command"""
        # Run subcommand in context
        subcommand = self._subcommand_dict[key]
        return ctx(subcommand, *values)

    # Controlled callbacks

    def _run_callback(self, ctx, func, node):
        """Contexualize different node callbacks."""
        try:
            with context(ctx, node):
                func(node)
        except Exception as exc:
            self.ignore_exception(exc)

    def _standard_aggregation(self, node, func, *nodes):
        """Contextualize aggregation and propagate errors automatically."""
        # Forward first exception
        for subnode in nodes:
            if subnode.exception() is not None:
                with context("updating", node):
                    raise subnode.exception()
        # Shortcut for empty nodes
        results = [subnode.result() for subnode in nodes]
        if any(result is None for result in results):
            return
        # Exctract values
        values, stamps, qualities = zip(*results)
        # Invalid quality
        if any(quality == INVALID for quality in qualities):
            return triplet(None, max(stamps), INVALID)
        # Run function
        try:
            with context("updating", node):
                result = func(*values)
        except Exception as exc:
            self.ignore_exception(exc)
            raise exc
        # Return triplet
        if isinstance(result, triplet):
            return result
        # Create triplet
        quality = aggregate_qualities(qualities)
        return triplet(result, max(stamps), quality)

    def _custom_aggregation(self, node, func, *nodes):
        """Contextualize aggregation."""
        # Run function
        try:
            with context("updating", node):
                result = func(*nodes)
        except Exception as exc:
            self.ignore_exception(exc)
            raise exc
        # Return result
        if not isinstance(result, triplet):
            result = triplet(result)
        return result

    # Dedicated callbacks

    def _set_state_from_node(self, node):
        # Forward exception
        if node.exception() is not None:
            self.register_exception(node.exception())
            return
        # Empty node
        result = node.result()
        if result is None:
            self.set_state(DevState.UNKNOWN)
            self.set_status("The state is currently not available.")
            return
        # Unpack triplet
        value, stamp, quality = result
        # Invalid value
        if value is None:
            value = (
                DevState.FAULT,
                "The state cannot be computed. Some values are invalid.",
            )
        # Unpack value
        try:
            state, status = value
        except (TypeError, ValueError):
            state = value
            status = f"The device is in {value} state."
        # Set state and status
        try:
            with context("setting", "state and status"):
                self.set_state(state, stamp, quality)
                self.set_status(status, stamp, quality)
        # Exception while setting the state
        except Exception as exc:
            self.register_exception(exc)

    def _push_event_for_node(self, node):
        # Exception
        if node.exception() is not None:
            exception = to_dev_failed(node.exception())
            self.push_change_event(node.name, exception)
            self.push_archive_event(node.name, exception)
            # Log the pushing of exceptions
            msg = "Pushing an exception for attribute {}"
            self.debug_exception(exception, msg.format(node.name))
        # Empty result
        elif node.result() is None:
            pass
        # Triplet result
        else:
            value, stamp, quality = node.result()
            if value is None:
                value = node.value_for_none
            self.push_change_event(node.name, value, stamp, quality)
            self.push_archive_event(node.name, value, stamp, quality)

    # Clean up

    def delete_device(self):
        # Reset graph
        try:
            self._graph.reset()
        except Exception as exc:
            msg = "Error while resetting the graph"
            self.ignore_exception(exc, msg)
        # Unsubscribe all
        super().delete_device()


# Timed Facade


class TimedFacade(Facade):
    """Similar to the `facadedevice.Facade` base class with time handling.

    In particular, it adds:

    - the `UpdateTime` polled command, used trigger updates periodically
    - the `Time` local attribute, a float updated at every tick
    - the `on_time` method, a callback that runs at every tick
    """

    def init_device(self):
        super().init_device()
        self.UpdateTime()

    Time = local_attribute(dtype=float)

    @Time.notify
    def _on_time(self, node):
        self.on_time(node.result().value)

    def on_time(self, value):
        pass

    @command(polling_period=1000, display_level=DispLevel.EXPERT)
    def UpdateTime(self):
        t = timestamp_now()
        result = triplet(t, t)
        self.graph["Time"].set_result(result)



"""Provide helpers for exception handling."""

import sys
import traceback
from contextlib import contextmanager

from tango import DevFailed, Except

# Safe traceback string


def traceback_string(exc, limit=None):
    if getattr(exc, "__traceback__", None):
        return "".join(traceback.format_tb(exc.__traceback__, limit=limit))
    if any(sys.exc_info()):
        return traceback.format_exc(limit=limit)
    return ""


# Safe exception representation


def exception_string(exc, wrap=None):
    # Convert DevFailed
    if isinstance(exc, DevFailed) and exc.args:
        exc = exc.args[0]
    # Exception as a string
    try:
        base = exc.desc
    except AttributeError:
        base = str(exc) if str(exc) else repr(exc)
    # No wrapping
    if not wrap:
        return base
    # Wrapping
    indented = "\n".join("  " + line for line in base.splitlines())
    return f"{wrap}:\n{indented}"


# DevFailed conversion


def to_dev_failed(exc):
    tb = traceback_string(exc)
    desc = exception_string(exc)
    try:
        Except.throw_exception("PyDs_PythonError", desc, tb)
    except Exception as exc:
        return exc


# Exception context


class ContextException(Exception):
    def __init__(self, base, context, origin, traceback):
        super().__init__(base, context, origin)
        self.base = base
        self.context = context
        self.origin = origin
        self.__traceback__ = traceback

    @property
    def desc(self):
        wrap = f"Exception while {self.context} {self.origin}"
        return exception_string(self.base, wrap)

    def __str__(self):
        return self.desc


# Exception context manager


@contextmanager
def context(msg, origin):
    try:
        yield
    except Exception as exc:
        _, _, tb = sys.exc_info()
        raise ContextException(exc, msg, origin, tb)



from tango import DevState

from facadedevice import Facade


class On(Facade):
    def safe_init_device(self):
        super().safe_init_device()
        self.set_state(DevState.ON)


if __name__ == "__main__":
    On.run_server()



from facadedevice import Facade, combined_attribute


class Average(Facade):
    @combined_attribute(dtype=float, property_name="AttributesToAverage")
    def average(self, *args):
        return sum(args) / len(args)


if __name__ == "__main__":
    Average.run_server()



from facadevice import Facade, logical_attribute, proxy_attribute


class Rectangle(Facade):
    Width = proxy_attribute(property_name="WidthAttribute")

    Height = proxy_attribute(property_name="HeightAttribute")

    @logical_attribute(bind=["Width", "Height"])
    def Area(width, height):
        return width * height


if __name__ == "__main__":
    Rectangle.run_server()



"""Provide a proxy device to subclass."""

from facadedevice.device import Facade, TimedFacade
from facadedevice.graph import triplet
from facadedevice.objects import (
    combined_attribute,
    local_attribute,
    logical_attribute,
    proxy_attribute,
    proxy_command,
    state_attribute,
)

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "0.0+unknown"

__all__ = [
    "Facade",
    "TimedFacade",
    "triplet",
    "proxy_attribute",
    "local_attribute",
    "logical_attribute",
    "state_attribute",
    "proxy_command",
    "combined_attribute",
]
